#!/usr/bin/env python2.6
# -*- coding: utf-8 -*-
# This software is distributed under the two-clause BSD license.

__version__ = '0.3.8'
__author__ = "Raphaël Barrois <raphael.barrois+uconf@polytechnique.org>"
