# -*- coding: utf-8 -*-
# Copyright (c) 2010-2013 Raphaël Barrois
# This software is distributed under the two-clause BSD license.


from __future__ import unicode_literals, absolute_import


"""config.h-style file holding all global constants for uconf."""


CONFIG_FILES = ('/etc/uconf.conf', '~/.uconfrc')
REPO_SUBFOLDER = '.uconf'
